﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Timers;
using Ozeki.Media;
using Ozeki.VoIP;

namespace CallQueueStrategies.CallQueue
{
    class CallQueueCallHandler
    {
        ICall call;
        Timer timer;
        Timer timerAgent;
        ICallManager callManager;
        List<string> members;

        MediaConnector mediaConnector;
        PhoneCallAudioSender phoneCallAudioSender;
        MP3StreamPlayback mp3Player;
        TextToSpeech textToSpeech;

        CallQueueHandlerContainer callQueueHandlerContainer;
        double waitingTime;
        private CallQueueStrategy strategy;
        string lastAgent;

        public CallQueueCallHandler(ICall call, ICallManager callManager, List<string> members, CallQueueHandlerContainer callQueueHandlerContainer, CallQueueStrategy strategy)
        {
            timer = new Timer(1000);
            timer.AutoReset = true;
            timer.Elapsed += timer_Elapsed;

            timerAgent = new Timer(15000);
            timerAgent.AutoReset = true;
            timerAgent.Elapsed += timerAgent_Elapsed;

            this.members = members;
            this.callManager = callManager;
            this.callQueueHandlerContainer = callQueueHandlerContainer;

            this.call = call;
            this.call.CallStateChanged += call_CallStateChanged;
            this.strategy = strategy;

            mediaConnector = new MediaConnector();
            phoneCallAudioSender = new PhoneCallAudioSender();
            mp3Player = new MP3StreamPlayback(@"test.mp3", true, true);
            phoneCallAudioSender.AttachToCall(call);

            textToSpeech = new TextToSpeech();
            textToSpeech.Stopped += TextToSpeechCompleted;

            mediaConnector.Connect(mp3Player, phoneCallAudioSender);
            mediaConnector.Connect(textToSpeech, phoneCallAudioSender);

            timer.Start();
            timerAgent.Start();
        }

        public string SelectedAgent { get; private set; }

        public void Start(string lastAgent = null)
        {
            this.lastAgent = lastAgent;

            if (call.CallState.IsCallEnded())
            {
                OnCompleted();
                return;
            }
        }

        private void OnCompleted()
        {
            mediaConnector.Dispose();
            mp3Player.Dispose();
            textToSpeech.Dispose();

            var handler = Completed;
            if (handler != null)
                handler(this, EventArgs.Empty);
        }

        public event EventHandler Completed;

        void timer_Elapsed(object sender, ElapsedEventArgs e)
        {
            timer.Stop();
            if (!call.CallState.IsInCall())
            {
                timer.Start();
                return;
            }

            var memb = new List<string>(members);

            foreach (var activeSession in callManager.ActiveSessions)
            {
                var callee = activeSession.CalleeInfo.Owner.ExtensionID;
                var caller = activeSession.CallerInfo.Owner.ExtensionID;

                memb.Remove(callee);
                memb.Remove(caller);

                double sum = 0;
                var callCounter = 0;

                foreach (var m in memb)
                {
                    List<ISession> historyResult = CallHistory.GetCallHistory(m);

                    foreach (var session in historyResult)
                    {
                        ++callCounter;
                        sum += session.TalkDuration.TotalSeconds;
                    }
                }

                if (sum != 0)
                {
                    waitingTime = sum / callCounter;
                }
            }

            SelectedAgent = GetAgent(memb);
                if (SelectedAgent != null)
                {
                    timerAgent.Stop();
                    call.BlindTransfer(SelectedAgent);
                }
           
        }

        private string GetAgent(List<string> memb)
        {
            switch (strategy)
            {
                case CallQueueStrategy.MostIdleAgent:
                    return GetMostIdleAgent(memb);
                    break;
                case CallQueueStrategy.LeastUtilizedAgent:
                    return GetLeastUtilizedAgent(memb);
                    break;
                case CallQueueStrategy.RoundRobin:
                    return GetRoundRobin(memb);
                    break;
            }
            return null;
        }

        private string GetRoundRobin(List<string> memb)
        {
            string selectedMember = null;
            for (int i = 0; i < memb.Count; i++)
            {
                if (lastAgent == null)
                {
                    selectedMember = memb[0];
                }
                else if (memb[i] == lastAgent)
                {
                    if (memb[i] == memb.LastOrDefault())
                    {
                        selectedMember = memb[0];
                    }
                    else
                    {
                        selectedMember = memb[i + 1];
                    }
                }
            }
            return selectedMember;
        }

        private string GetMostIdleAgent(List<string> memb)
        {
            double tmp = 0;
            double idleTime = 0;
            string selectedMember = null;

            foreach (var m in memb)
            {
                List<ISession> historyResult = CallHistory.GetCallHistory(m);
                var lastHistoryItem = historyResult.LastOrDefault();

                if (lastHistoryItem == null)
                {
                    return m;
                }

                idleTime = (DateTime.Now.Hour * 3600 + DateTime.Now.Minute * 60 + DateTime.Now.Second) -
                           (lastHistoryItem.StartTime.Hour * 3600 + lastHistoryItem.StartTime.Minute * 60 + lastHistoryItem.StartTime.Second +
                            lastHistoryItem.TalkDuration.TotalSeconds);

                if (idleTime != 0 && idleTime > tmp)
                {
                    tmp = idleTime;
                    selectedMember = m;
                }
            }
            return selectedMember;
        }

        private string GetLeastUtilizedAgent(List<string> memb)
        {
            double inCallTime = 0;
            double idleTime = 0;
            double currentUtilized = 0;
            string selectedMember = null;

            for (int i = 0; i < memb.Count; i++)
            {
                List<ISession> historyResult = CallHistory.GetCallHistory(memb[i]);
                var currentTime = DateTime.Now;
                var firstHistoryItem = historyResult.FirstOrDefault(item => item.StartTime.Year == currentTime.Year 
                                                                    && item.StartTime.Month == currentTime.Month 
                                                                    && item.StartTime.Day == currentTime.Day); //

                TimeSpan diff = (currentTime - new DateTime(currentTime.Year, currentTime.Month, currentTime.Day));
                var spendTime = diff.TotalSeconds;

                if (firstHistoryItem == null)
                {
                    return memb[i];
                }

                foreach (var session in historyResult)
                {
                    inCallTime = inCallTime + session.TalkDuration.TotalSeconds;
                }

                idleTime = spendTime - inCallTime;
                double utilized = inCallTime / idleTime;

                if (i == 0)
                {
                    currentUtilized = utilized;
                    selectedMember = memb[0];
                }
                else if (utilized < currentUtilized)
                {
                    currentUtilized = utilized;
                    selectedMember = memb[i];
                }
            }
            return selectedMember;
        }


        void timerAgent_Elapsed(object sender, ElapsedEventArgs e)
        {
            mp3Player.Pause();
            textToSpeech.Clear();

            if (waitingTime == 0)
            {
                textToSpeech.AddAndStartText(string.Format("{0} more customers are waiting in line before you. Please be patient.",
                                                           callQueueHandlerContainer.GetIndex(this)));
            }
            else
            {
                textToSpeech.AddAndStartText(string.Format("{0} more customers are waiting in line before you. " +
                                                           "Your estimated wait time is approximately {1} seconds. Please be patient.",
                                                           callQueueHandlerContainer.GetIndex(this),
                                                           (int)waitingTime));
            }
        }

        void call_CallStateChanged(object sender, CallStateChangedArgs e)
        {
            if (e.State == CallState.Answered)
            {
                mp3Player.Start();
            }

            if (e.State.IsRinging())
            {
                call.Answer();
            }

            else if (e.State.IsCallEnded())
            {
                OnCompleted();
            }
        }

        void TextToSpeechCompleted(object sender, EventArgs e)
        {
            mp3Player.Start();
            textToSpeech.Clear();
        }
    }
}
